#!/usr/bin/env python3
# Copyright (C) <2018-2021>  <Agence Data Services, DSI Pôle Emploi>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import shutil
import numpy as np
import pandas as pd
import dill as pickle

from sklearn.impute import SimpleImputer
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.feature_selection import SelectKBest, SelectorMixin
from sklearn.feature_extraction.text import CountVectorizer, _VectorizerMixin
from sklearn.compose import ColumnTransformer, make_column_transformer, make_column_selector
from sklearn.preprocessing import FunctionTransformer, StandardScaler, MinMaxScaler, KBinsDiscretizer, Binarizer, PolynomialFeatures, OneHotEncoder

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.models_training import utils_models
from {{package_name}}.models_training.classifiers import model_rf_classifier, model_dense_classifier, model_xgboost_classifier
from {{package_name}}.models_training.regressors import model_rf_regressor, model_dense_regressor, model_xgboost_regressor


# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class UtilsModelsTests(unittest.TestCase):
    '''Main class to test all functions in utils_models.py'''

    # We avoid tqdm prints
    pd.Series.progress_apply = pd.Series.apply


    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)


    def test01_normal_split(self):
        '''Test of the function {{package_name}}.models_training.utils_models.normal_split'''
        # Valids to test
        input_test = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']})
        test_size = 0.2
        train, test = utils_models.normal_split(input_test, test_size=test_size)
        self.assertEqual(train.shape[0], input_test.shape[0] * (1 - test_size))
        self.assertEqual(test.shape[0], input_test.shape[0] * test_size)

        # Check inputs
        with self.assertRaises(ValueError):
            utils_models.normal_split(input_test, test_size=1.2)
        with self.assertRaises(ValueError):
            utils_models.normal_split(input_test, test_size=-0.2)


    def test02_stratified_split(self):
        '''Test of the function {{package_name}}.models_training.utils_models.stratified_split'''
        # Valids to test
        input_test = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k'],
                                   'col2': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2]})
        test_size = 0.5
        col = 'col2'
        train, test = utils_models.stratified_split(input_test, col, test_size=test_size)
        self.assertEqual(train.shape[0], (input_test.shape[0] - 1) * (1 - test_size))
        self.assertEqual(test.shape[0], (input_test.shape[0] - 1) * test_size)
        self.assertEqual(train[train[col] == 0].shape[0], input_test[input_test[col] == 0].shape[0] * (1 - test_size))
        self.assertEqual(test[test[col] == 0].shape[0], input_test[input_test[col] == 0].shape[0] * test_size)
        self.assertEqual(train[train[col] == 2].shape[0], 0)
        self.assertEqual(test[test[col] == 2].shape[0], 0)

        # Check the input(s) type(s)
        with self.assertRaises(ValueError):
            utils_models.stratified_split(input_test, col, test_size=1.2)
        with self.assertRaises(ValueError):
            utils_models.stratified_split(input_test, col, test_size=-0.2)


    def test03_remove_small_classes(self):
        '''Test of the function {{package_name}}.models_training.utils_models.remove_small_classes'''
        # Valids to test
        input_test = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k'],
                                   'col2': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2]})
        test_size = 0.2
        col = 'col2'

        result_df = utils_models.remove_small_classes(input_test, col, min_rows=2)
        self.assertEqual(result_df[result_df[col] == 0].shape[0], input_test[input_test[col] == 0].shape[0])
        self.assertEqual(result_df[result_df[col] == 1].shape[0], input_test[input_test[col] == 1].shape[0])
        self.assertEqual(result_df[result_df[col] == 2].shape[0], 0)

        result_df = utils_models.remove_small_classes(input_test, col, min_rows=5)
        self.assertEqual(result_df[result_df[col] == 0].shape[0], 0)
        self.assertEqual(result_df[result_df[col] == 1].shape[0], input_test[input_test[col] == 1].shape[0])
        self.assertEqual(result_df[result_df[col] == 2].shape[0], 0)

        # Check the input(s) type(s)
        with self.assertRaises(ValueError):
            utils_models.remove_small_classes(input_test, col, min_rows=0)


    def test04_display_train_test_shape(self):
        '''Test of the function {{package_name}}.models_training.utils_models.display_train_test_shape'''
        # Valids to test
        df = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k'],
                           'col2': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2]})

        # Nominal case
        utils_models.display_train_test_shape(df, df)
        utils_models.display_train_test_shape(df, df, df_shape=10)


    def test05_preprocess_model_multilabel(self):
        '''Test of the function {{package_name}}.models_training.utils_models.preprocess_model_multilabel'''
        # Create a dataset
        df = pd.DataFrame({'x_col': ['test', 'toto', 'titi'], 'y_col': [(), ('x1', 'x2'), ('x3', 'x4', 'x1')]})
        df_expected = pd.DataFrame({'x_col': ['test', 'toto', 'titi'], 'y_col': [(), ('x1', 'x2'), ('x3', 'x4', 'x1')],
                                    'x1': [0, 1, 1], 'x2': [0, 1, 0], 'x3': [0, 0, 1], 'x4': [0, 0, 1]})
        subset_classes = ['x2', 'x4']
        df_subset_expected = pd.DataFrame({'x_col': ['test', 'toto', 'titi'], 'y_col': [(), ('x1', 'x2'), ('x3', 'x4', 'x1')],
                                           'x2': [0, 1, 0], 'x4': [0, 0, 1]})

        # Nominal case
        df_mlb, classes = utils_models.preprocess_model_multilabel(df, 'y_col')
        self.assertEqual(sorted(classes), ['x1', 'x2', 'x3', 'x4'])
        pd.testing.assert_frame_equal(df_mlb, df_expected, check_dtype=False)

        # Test argument classes
        df_mlb, classes = utils_models.preprocess_model_multilabel(df, 'y_col', classes=subset_classes)
        self.assertEqual(sorted(classes), sorted(subset_classes))
        pd.testing.assert_frame_equal(df_mlb, df_subset_expected, check_dtype=False)

    def test06_load_pipeline(self):
        '''Test of the function {{package_name}}.models_training.utils_models.load_pipeline'''

        # Creation mock pipeline
        pipeline_dir = os.path.join(utils.get_pipelines_path(), 'fake_pipeline_dir')
        pipeline_path = os.path.join(pipeline_dir, 'pipeline.pkl')
        preprocess_str = 'fake_pipeline'
        remove_dir(pipeline_dir)
        os.makedirs(pipeline_dir)
        fake_pipeline = ColumnTransformer([('fake_pipeline', FunctionTransformer(lambda x: x * 2), ['toto', 'titi'])])
        df = pd.DataFrame({'toto': [1, 2, 3], 'tata': [4, 5, 6], 'titi': [7, 8, 9]})
        fake_pipeline.fit(df) # We fit even if it is not really necessary

        # Save pipeline
        pipeline_dict = {'preprocess_pipeline': fake_pipeline, 'preprocess_str': preprocess_str}
        with open(pipeline_path, 'wb') as f:
            pickle.dump(pipeline_dict, f)

        # Reload
        new_pipeline, new_preprocess_str = utils_models.load_pipeline(pipeline_dir='fake_pipeline_dir')
        # We perform some tests
        np.testing.assert_equal(new_pipeline.transform(df), fake_pipeline.transform(df))
        self.assertEqual(new_preprocess_str, preprocess_str)

        # Same thing but with a path
        new_pipeline, new_preprocess_str = utils_models.load_pipeline(pipeline_dir=pipeline_dir, is_path=True)
        # We perform some tests
        np.testing.assert_equal(new_pipeline.transform(df), fake_pipeline.transform(df))
        self.assertEqual(new_preprocess_str, preprocess_str)
        remove_dir(pipeline_dir)

        # We do the same thing with pipeline_dir=None, ie. backup no_preprocess
        new_pipeline, new_preprocess_str = utils_models.load_pipeline(pipeline_dir=None)
        # Warning, we must fit the pipeline
        new_pipeline.fit(df)
        # We perform some tests
        np.testing.assert_equal(new_pipeline.transform(df), np.array(df))
        self.assertEqual(new_preprocess_str, 'no_preprocess')

        # Check errors
        with self.assertRaises(FileNotFoundError):
            utils_models.load_pipeline(pipeline_dir='tototo')
        with self.assertRaises(FileNotFoundError):
            utils_models.load_pipeline(pipeline_dir='./tototo', is_path=True)
        with self.assertRaises(NotFittedError):
            pipeline, _ = utils_models.load_pipeline(pipeline_dir=None)
            pipeline.transform(df)


    def test07_load_model(self):
        '''Test of the function {{package_name}}.models_training.utils_models.load_model'''

        # Data for training
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, 2, -6, 3], 'col_2': [2, -1, -8, 3, 12, 2]})
        x_test = pd.DataFrame({'col_1': [-2, 8, 0, -1, -2, -9], 'col_2': [5, -1, 3, 6, -5, 2]})
        y_train_classification = pd.Series([0, 0, 0, 1, 1, 1])
        y_train_regression = pd.Series([-3, -2, -8, 5, 6, 5])
        model_dir = os.path.join(utils.get_models_path(), 'test_model')
        model_name = 'test_model_name'
        batch_size = 8
        epochs = 3
        patience = 5
        early_stopping_rounds = 3

        ####################################################

        # Tests on a mock model - RF classifier
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(x_train, y_train_classification)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        remove_dir(model_dir)

        ####################################################

        # Tests on a mock model - RF regressor
        remove_dir(model_dir)
        model = model_rf_regressor.ModelRFRegressor(model_dir=model_dir, model_name=model_name)
        model.fit(x_train, y_train_regression)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        remove_dir(model_dir)

        ####################################################

        # Tests on a mock model - Dense classifier
        remove_dir(model_dir)
        model = model_dense_classifier.ModelDenseClassifier(model_dir=model_dir, model_name=model_name,
                                                            batch_size=batch_size, epochs=epochs, patience=patience)
        model.fit(x_train, y_train_classification)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['batch_size'], batch_size)
        self.assertEqual(new_config['epochs'], epochs)
        self.assertEqual(new_config['patience'], patience)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        self.assertEqual([list(_) for _ in new_model.predict_proba(x_test)], [list(_) for _ in model.predict_proba(x_test)])

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['batch_size'], batch_size)
        self.assertEqual(new_config['epochs'], epochs)
        self.assertEqual(new_config['patience'], patience)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        self.assertEqual([list(_) for _ in new_model.predict_proba(x_test)], [list(_) for _ in model.predict_proba(x_test)])
        remove_dir(model_dir)

        ####################################################

        # Tests on a mock model - Dense regressor
        remove_dir(model_dir)
        model = model_dense_regressor.ModelDenseRegressor(model_dir=model_dir, model_name=model_name,
                                                          batch_size=batch_size, epochs=epochs, patience=patience)
        model.fit(x_train, y_train_regression)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['batch_size'], batch_size)
        self.assertEqual(new_config['epochs'], epochs)
        self.assertEqual(new_config['patience'], patience)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['batch_size'], batch_size)
        self.assertEqual(new_config['epochs'], epochs)
        self.assertEqual(new_config['patience'], patience)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        remove_dir(model_dir)

        ####################################################

        # Tests on a mock model - XGboost classifier
        remove_dir(model_dir)
        model = model_xgboost_classifier.ModelXgboostClassifier(model_dir=model_dir, model_name=model_name,
                                                                xgboost_params={'n_estimators': 5},
                                                                early_stopping_rounds=early_stopping_rounds)
        model.fit(x_train, y_train_classification)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['early_stopping_rounds'], early_stopping_rounds)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        self.assertEqual([list(_) for _ in new_model.predict_proba(x_test)], [list(_) for _ in model.predict_proba(x_test)])

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['early_stopping_rounds'], early_stopping_rounds)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        self.assertEqual([list(_) for _ in new_model.predict_proba(x_test)], [list(_) for _ in model.predict_proba(x_test)])
        remove_dir(model_dir)

        ####################################################

        # Tests on a mock model - XGboost regressor
        remove_dir(model_dir)
        model = model_xgboost_regressor.ModelXgboostRegressor(model_dir=model_dir, model_name=model_name,
                                                              xgboost_params={'n_estimators': 5},
                                                              early_stopping_rounds=early_stopping_rounds)
        model.fit(x_train, y_train_regression)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['early_stopping_rounds'], early_stopping_rounds)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['early_stopping_rounds'], early_stopping_rounds)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(x_test)), list(model.predict(x_test)))
        remove_dir(model_dir)

        ####################################################

        # Check errors
        with self.assertRaises(FileNotFoundError):
            utils_models.load_model(model_dir='tototo')
        with self.assertRaises(FileNotFoundError):
            utils_models.load_model(model_dir='./tototo', is_path=True)


    def test08_get_columns_pipeline(self):
        '''Test of the function {{package_name}}.models_training.utils_models.get_columns_pipeline'''
        # Set fake pipelines
        numeric_pipeline = make_pipeline(SimpleImputer(strategy='median'), StandardScaler())
        cat_pipeline = make_pipeline(SimpleImputer(strategy='most_frequent'), OneHotEncoder(handle_unknown='ignore'))
        text_pipeline = make_pipeline(CountVectorizer(), SelectKBest(k=5))
        transformers = [
            ('num', numeric_pipeline, ['Age', 'SibSp', 'Parch', 'Fare']),
            ('cat', cat_pipeline, ['Pclass', 'Sex', 'Embarked']), # To cast a column to 'category' -> df["A"].astype("category")
            ('text', text_pipeline, 'Name'), # CountVectorizer possible one column at a time
        ]
        pipeline1 = ColumnTransformer(transformers, remainder='drop')
        pipeline2 = ColumnTransformer(transformers, remainder='passthrough')

        # Fake DataFrame to fit the pipelines
        df = pd.read_csv('test_dataset.csv', sep=',', encoding='utf-8')
        y = df['Survived']
        X = df.drop('Survived', axis=1)
        pipeline1.fit(X, y)
        pipeline2.fit(X, y)

        # Nominal case
        columns_in1, mandatory_columns1 = utils_models.get_columns_pipeline(pipeline1)
        columns_in2, mandatory_columns2 = utils_models.get_columns_pipeline(pipeline2)
        self.assertEqual(set(['PassengerId', 'Pclass', 'Name', 'Sex', 'Age', 'SibSp', 'Parch', 'Ticket', 'Fare', 'Cabin', 'Embarked']), set(columns_in1))
        self.assertEqual(set(['Age', 'SibSp', 'Parch', 'Fare', 'Pclass', 'Sex', 'Embarked', 'Name']), set(mandatory_columns1))
        self.assertEqual(set(['PassengerId', 'Pclass', 'Name', 'Sex', 'Age', 'SibSp', 'Parch', 'Ticket', 'Fare', 'Cabin', 'Embarked']), set(columns_in2))
        self.assertEqual(set(['PassengerId', 'Pclass', 'Name', 'Sex', 'Age', 'SibSp', 'Parch', 'Ticket', 'Fare', 'Cabin', 'Embarked']), set(mandatory_columns2))


    def test09_apply_pipeline(self):
        '''Test of the function {{package_name}}.models_training.utils_models.apply_pipeline'''
        # Set fake pipelines
        numeric_pipeline = make_pipeline(SimpleImputer(strategy='median'), StandardScaler())
        cat_pipeline = make_pipeline(SimpleImputer(strategy='most_frequent'), OneHotEncoder(handle_unknown='ignore'))
        text_pipeline = make_pipeline(CountVectorizer(), SelectKBest(k=5))
        transformers = [
            ('num', numeric_pipeline, ['Age', 'SibSp', 'Parch', 'Fare']),
            ('cat', cat_pipeline, ['Pclass', 'Sex', 'Embarked']), # To cast a column to 'category' -> df["A"].astype("category")
            ('text', text_pipeline, 'Name'), # CountVectorizer possible one column at a time
        ]
        pipeline1 = ColumnTransformer(transformers, remainder='drop')
        pipeline2 = ColumnTransformer(transformers, remainder='passthrough')

        # Fake DataFrame to fit the pipelines
        df = pd.read_csv('test_dataset.csv', sep=',', encoding='utf-8')
        y = df['Survived']
        X = df.drop('Survived', axis=1)
        pipeline1.fit(X, y)
        pipeline2.fit(X, y)

        # Nominal case
        preprocessed_df1 = utils_models.apply_pipeline(X, pipeline1)
        self.assertEqual(preprocessed_df1.shape, (100, 17))
        preprocessed_df2 = utils_models.apply_pipeline(X, pipeline2)
        self.assertEqual(preprocessed_df2.shape, (100, 20))

        # With useless columns
        df['toto'] = 5
        preprocessed_df1 = utils_models.apply_pipeline(df, pipeline1)
        self.assertEqual(preprocessed_df1.shape, (100, 17))
        preprocessed_df2 = utils_models.apply_pipeline(df, pipeline2)
        self.assertEqual(preprocessed_df2.shape, (100, 20))

        # With missing optional column
        new_X = X.drop('PassengerId', axis=1)
        preprocessed_df1 = utils_models.apply_pipeline(new_X, pipeline1)
        self.assertEqual(preprocessed_df1.shape, (100, 17))
        with self.assertRaises(ValueError):
            preprocessed_df2 = utils_models.apply_pipeline(new_X, pipeline2)

        # With missing mandatory column
        new_X = X.drop('Age', axis=1)
        with self.assertRaises(ValueError):
            preprocessed_df1 = utils_models.apply_pipeline(new_X, pipeline1)
        with self.assertRaises(ValueError):
            preprocessed_df2 = utils_models.apply_pipeline(new_X, pipeline2)


    def test10_predict(self):
        '''Test of the function {{package_name}}.models_training.utils_models.predict'''

        # Data for training
        x_train_1 = pd.DataFrame({'col_1': [-5, 5, -5, 5, 5, 5] * 10})
        x_train_2 = pd.DataFrame({'col_1': [-5, -5, -5, 5, 5, 5] * 10, 'col_2': [0, 0, 0, 0, 0, 0] * 10})
        x_test_1 = pd.DataFrame({'col_1': [-5, 5, -5, 5, -5, 5]})
        x_test_1_solo = pd.DataFrame({'col_1': [-5]})
        x_test_2 = pd.DataFrame({'col_1': [-5, 5, -5, 5, -5, 5], 'col_2': [0, 0, 0, 0, 0, 0]})
        x_test_2_solo = pd.DataFrame({'col_1': [5], 'col_2': [0]})
        y_train_classification = pd.Series([0, 0, 0, 1, 1, 1] * 10)
        y_train_classification_multi = pd.DataFrame({'test': [1, 1, 1, 0, 0, 0] * 10, 'toto': [0, 0, 0, 1, 1, 1] * 10})
        y_train_regression = pd.Series([-10, -10, -10, 10, 10, 10] * 10)
        y_test_classification = [0, 1, 0, 1, 0, 1]
        y_test_classification_multi = [('test',), ('toto',), ('test',), ('toto',), ('test',), ('toto',)]
        y_test_1_classification_solo = 0
        y_test_2_classification_solo = 1
        y_test_1_classification_multi_solo = ('test',)
        y_test_2_classification_multi_solo = ('toto',)
        model_dir = os.path.join(utils.get_models_path(), 'test_model')
        model_name = 'test_model_name'

        ################
        # RF Classification - mono-label - 1

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_1, y_train_classification)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        self.assertEqual(utils_models.predict(x_test_1, model), y_test_classification)
        self.assertEqual(utils_models.predict(x_test_1_solo, model), y_test_1_classification_solo)
        remove_dir(model_dir)

        ################
        # RF Classification - mono-label - 2

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_2, y_train_classification)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        self.assertEqual(utils_models.predict(x_test_2, model), y_test_classification)
        self.assertEqual(utils_models.predict(x_test_2_solo, model), y_test_2_classification_solo)
        remove_dir(model_dir)

        ################
        # RF Classification - multi-labels - 1

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name, multi_label=True)
        model.fit(x_train_1, y_train_classification_multi)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        self.assertEqual(utils_models.predict(x_test_1, model), y_test_classification_multi)
        self.assertEqual(utils_models.predict(x_test_1_solo, model), y_test_1_classification_multi_solo)
        remove_dir(model_dir)

        ################
        # RF Classification - multi-labels - 2

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name, multi_label=True)
        model.fit(x_train_2, y_train_classification_multi)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        self.assertEqual(utils_models.predict(x_test_2, model), y_test_classification_multi)
        self.assertEqual(utils_models.predict(x_test_2_solo, model), y_test_2_classification_multi_solo)
        remove_dir(model_dir)

        ################
        # RF Regression - 1

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_regressor.ModelRFRegressor(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_1, y_train_regression)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        # Check just the type, regression is complicated...
        self.assertEqual(type(utils_models.predict(x_test_1, model)), list)
        self.assertTrue(isinstance(utils_models.predict(x_test_1_solo, model), (np.floating, float)))
        remove_dir(model_dir)

        ################
        # RF Regression - 2

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_regressor.ModelRFRegressor(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_2, y_train_regression)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        # Check just the type, regression is complicated...
        self.assertEqual(type(utils_models.predict(x_test_2, model)), list)
        self.assertTrue(isinstance(utils_models.predict(x_test_2_solo, model), (np.floating, float)))
        remove_dir(model_dir)


    def test11_predict_with_proba(self):
        '''Test of the function {{package_name}}.models_training.utils_models.predict_with_proba'''

        # Data for training
        x_train_1 = pd.DataFrame({'col_1': [-5, 5, -5, 5, 5, 5] * 10})
        x_train_2 = pd.DataFrame({'col_1': [-5, -5, -5, 5, 5, 5] * 10, 'col_2': [0, 0, 0, 0, 0, 0] * 10})
        x_test_1 = pd.DataFrame({'col_1': [-5, 5, -5, 5, -5, 5]})
        x_test_1_solo = pd.DataFrame({'col_1': [-5]})
        x_test_2 = pd.DataFrame({'col_1': [-5, 5, -5, 5, -5, 5], 'col_2': [0, 0, 0, 0, 0, 0]})
        x_test_2_solo = pd.DataFrame({'col_1': [5], 'col_2': [0]})
        y_train_classification = pd.Series([0, 0, 0, 1, 1, 1] * 10)
        y_train_classification_multi = pd.DataFrame({'test': [1, 1, 1, 0, 0, 0] * 10, 'toto': [0, 0, 0, 1, 1, 1] * 10})
        y_train_regression = pd.Series([-10, -10, -10, 10, 10, 10] * 10)
        y_test_classification = [0, 1, 0, 1, 0, 1]
        y_test_classification_multi = [('test',), ('toto',), ('test',), ('toto',), ('test',), ('toto',)]
        y_test_1_classification_solo = 0
        y_test_2_classification_solo = 1
        y_test_1_classification_multi_solo = ('test',)
        y_test_2_classification_multi_solo = ('toto',)
        model_dir = os.path.join(utils.get_models_path(), 'test_model')
        model_name = 'test_model_name'

        ################
        # RF Classification - mono-label - 1

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_1, y_train_classification)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        pred, proba = utils_models.predict_with_proba(x_test_1, model)
        self.assertEqual(pred, y_test_classification)
        self.assertEqual(len(proba), len(y_test_classification))
        self.assertEqual(sum([round(p) for p in proba]), len(y_test_classification))
        pred, proba = utils_models.predict_with_proba(x_test_1_solo, model)
        self.assertEqual(pred, y_test_1_classification_solo)
        self.assertTrue(proba >= 0.5)
        remove_dir(model_dir)

        ################
        # RF Classification - mono-label - 2

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_2, y_train_classification)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        pred, proba = utils_models.predict_with_proba(x_test_2, model)
        self.assertEqual(pred, y_test_classification)
        self.assertEqual(len(proba), len(y_test_classification))
        self.assertEqual(sum([round(p) for p in proba]), len(y_test_classification))
        pred, proba = utils_models.predict_with_proba(x_test_2_solo, model)
        self.assertEqual(pred, y_test_2_classification_solo)
        self.assertTrue(proba >= 0.5)
        remove_dir(model_dir)

        ################
        # RF Classification - multi-labels - 1

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name, multi_label=True)
        model.fit(x_train_1, y_train_classification_multi)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        pred, proba = utils_models.predict_with_proba(x_test_1, model)
        self.assertEqual(pred, y_test_classification_multi)
        self.assertEqual(len(proba), len(y_test_classification_multi))
        self.assertEqual(sum([round(p[0]) for p in proba]), sum([len(_) for _ in y_test_classification_multi]))
        pred, proba = utils_models.predict_with_proba(x_test_1_solo, model)
        self.assertEqual(pred, y_test_1_classification_multi_solo)
        self.assertTrue(proba[0] >= 0.5)
        remove_dir(model_dir)

        ################
        # RF Classification - multi-labels - 2

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_classifier.ModelRFClassifier(model_dir=model_dir, model_name=model_name, multi_label=True)
        model.fit(x_train_2, y_train_classification_multi)
        model.save()

        # Nominal case
        model, model_conf = utils_models.load_model(model_dir='test_model')
        pred, proba = utils_models.predict_with_proba(x_test_2, model)
        self.assertEqual(pred, y_test_classification_multi)
        self.assertEqual(len(proba), len(y_test_classification_multi))
        self.assertEqual(sum([round(p[0]) for p in proba]), sum([len(_) for _ in y_test_classification_multi]))
        pred, proba = utils_models.predict_with_proba(x_test_2_solo, model)
        self.assertEqual(pred, y_test_2_classification_multi_solo)
        self.assertTrue(proba[0] >= 0.5)
        remove_dir(model_dir)

        ################
        # RF Regression

        # Creation fake model
        remove_dir(model_dir)
        model = model_rf_regressor.ModelRFRegressor(model_dir=model_dir, model_name=model_name)
        model.fit(x_train_1, y_train_regression)
        model.save()

        with self.assertRaises(ValueError):
            pred, proba = utils_models.predict_with_proba(x_test_1, model)

        remove_dir(model_dir)


    def test12_search_hp_cv(self):
        '''Test of the function {{package_name}}.models_training.utils_models.search_hp_cv'''
        # Definition of the variables for the nominal case
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 3, 12, 2] * 10})
        y_train_mono = pd.Series([0, 0, 0, 1, 1, 1] * 10)
        y_train_multi = pd.DataFrame({'y1': [0, 0, 0, 1, 1, 1] * 10, 'y2': [1, 0, 0, 1, 1, 1] * 10})
        model_cls = model_rf_classifier.ModelRFClassifier
        model_params_mono = {'multi_label': False}
        model_params_multi = {'multi_label': True}
        hp_params = {'rf_params': [{'max_depth': 5, 'n_estimators': 50}, {'max_depth': 10, 'n_estimators': 25}], 'multiclass_strategy': ['ovo', 'ovr']}
        kwargs_fit_mono = {'x_train': x_train, 'y_train': y_train_mono}
        kwargs_fit_multi = {'x_train': x_train, 'y_train': y_train_multi}

        # Nominal case
        n_splits = 5
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, "accuracy", kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Multi-labels case
        n_splits = 3
        model = utils_models.search_hp_cv(model_cls, model_params_multi, hp_params, "f1", kwargs_fit_multi, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Check the various scoring functions
        n_splits = 2
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, "precision", kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        n_splits = 2
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, "recall", kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Custom scoring function
        def custom_func(test_dict: dict):
            return (test_dict['Precision'] + test_dict['Recall']) / 2
        n_splits = 2
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, custom_func, kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)


        # Check errors
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'toto', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, {'toto': True}, hp_params, 'accuracy', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', {'toto': True, 'y_train': y_train_mono}, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', {'toto': True, 'x_train': x_train}, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', {'toto': True, 'x_train': x_train}, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, {**model_params_mono, **{'toto': True}}, {**hp_params, **{'toto': [False]}}, 'accuracy', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, {'toto': [1, 2], 'titi': [3]}, 'accuracy', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', kwargs_fit_mono, n_splits=1)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# TODO : There are some tests missing, 

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import tempfile
import numpy as np
import pandas as pd
from PIL import Image
from skimage import io

from {{package_name}} import utils
from {{package_name}}.models_training.object_detectors import utils_object_detectors

# Disable logging
import logging
logging.disable(logging.CRITICAL)


class UtilsObjectDetectorTests(unittest.TestCase):
    '''Main class to test all functions in utils_object_detectors.py'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_draw_bboxes_from_file(self):
        '''Test of the function utils_object_detectors.draw_bboxes_from_file'''

        # Set vars
        input_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'apple.jpg')
        with Image.open(input_path) as im:
            w, h = im.size
        gt_bboxes = [{'x1': 2, 'y1': 2, 'x2': 4, 'y2': 4, 'class': 'test', 'toto': 8}, {'x1': 1, 'y1': 1, 'x2': 10, 'y2': 8, 'titi': 'tata'}]
        predicted_bboxes = [{'x1': 3, 'y1': 2, 'x2': 4, 'y2': 4, 'class': 'titi', 'toto': 0.1}, {'x1': 5, 'y1': 1, 'x2': 8, 'y2': 22, 'titi': 'tata'}]

        # Nominal case
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes_from_file(input_path, output_path=output_path, gt_bboxes=gt_bboxes, predicted_bboxes=predicted_bboxes)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)

        # Without output
        output_im = utils_object_detectors.draw_bboxes_from_file(input_path, output_path=None, gt_bboxes=gt_bboxes, predicted_bboxes=predicted_bboxes)
        self.assertTrue(type(output_im) == np.ndarray)
        self.assertEqual(output_im.shape[0], h)
        self.assertEqual(output_im.shape[1], w)

        # Without gt_bboxes
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes_from_file(input_path, output_path=output_path, gt_bboxes=None, predicted_bboxes=predicted_bboxes)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)

        # Without predicted_bboxes
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes_from_file(input_path, output_path=output_path, gt_bboxes=gt_bboxes, predicted_bboxes=None)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)

        # Without gt_bboxes & predicted_bboxes
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes_from_file(input_path, output_path=output_path, gt_bboxes=None, predicted_bboxes=None)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)

        # Check errors
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            fake_path = os.path.join(tmp_folder, 'fake.jpg')
            with self.assertRaises(FileNotFoundError):
                output_im = utils_object_detectors.draw_bboxes_from_file(fake_path, output_path=output_path, gt_bboxes=gt_bboxes, predicted_bboxes=predicted_bboxes)

    def test02_draw_bboxes(self):
        '''Test of the function utils_object_detectors.draw_bboxes'''

        # Set vars
        input_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'apple.jpg')
        input_img = io.imread(input_path)
        h, w, depth = input_img.shape[0], input_img.shape[1], input_img.shape[2]
        gt_bboxes = [{'x1': 2, 'y1': 2, 'x2': 4, 'y2': 4, 'class': 'test', 'toto': 8}, {'x1': 1, 'y1': 1, 'x2': 10, 'y2': 8, 'titi': 'tata'}]
        predicted_bboxes = [{'x1': 3, 'y1': 2, 'x2': 4, 'y2': 4, 'class': 'titi', 'toto': 0.1}, {'x1': 5, 'y1': 1, 'x2': 8, 'y2': 22, 'titi': 'tata'}]

        # Nominal case
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes(input_img, output_path=output_path, gt_bboxes=gt_bboxes, predicted_bboxes=predicted_bboxes)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)
            self.assertEqual(output_im.shape[2], depth)

        # Without output
        output_im = utils_object_detectors.draw_bboxes(input_img, output_path=None, gt_bboxes=gt_bboxes, predicted_bboxes=predicted_bboxes)
        self.assertTrue(type(output_im) == np.ndarray)
        self.assertEqual(output_im.shape[0], h)
        self.assertEqual(output_im.shape[1], w)
        self.assertEqual(output_im.shape[2], depth)

        # Without gt_bboxes
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes(input_img, output_path=output_path, gt_bboxes=None, predicted_bboxes=predicted_bboxes)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)
            self.assertEqual(output_im.shape[2], depth)

        # Without predicted_bboxes
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes(input_img, output_path=output_path, gt_bboxes=gt_bboxes, predicted_bboxes=None)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)
            self.assertEqual(output_im.shape[2], depth)

        # Without gt_bboxes & predicted_bboxes
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmp_folder:
            output_path = os.path.join(tmp_folder, 'tmp_image.png')
            output_im = utils_object_detectors.draw_bboxes(input_img, output_path=output_path, gt_bboxes=None, predicted_bboxes=None)
            self.assertTrue(os.path.exists(output_path))
            self.assertTrue(type(output_im) == np.ndarray)
            self.assertEqual(output_im.shape[0], h)
            self.assertEqual(output_im.shape[1], w)

        # Check errors
        with tempfile.TemporaryFile(dir=os.getcwd()) as tmp_file:
            fake_path = tmp_file.name
            with self.assertRaises(FileExistsError):
                output_im = utils_object_detectors.draw_bboxes(input_img, output_path=fake_path, gt_bboxes=gt_bboxes, predicted_bboxes=predicted_bboxes)

    def test03_draw_rectangle_from_bbox(self):
        '''Test of the function utils_object_detectors.draw_rectangle_from_bbox'''

        # Set vars
        input_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'apple.jpg')
        input_img = io.imread(input_path)
        bbox_with_class = {'x1': 2, 'y1': 2, 'x2': 4, 'y2': 4, 'class': 'test', 'toto': 8}
        bbox_without_class = {'x1': 2, 'y1': 2, 'x2': 4, 'y2': 4}

        # Nominal case - we only test that the function does not give an error
        utils_object_detectors.draw_rectangle_from_bbox(input_img, bbox_with_class, color=None, thickness=None, with_center=False)
        # Without class
        utils_object_detectors.draw_rectangle_from_bbox(input_img, bbox_without_class, color=None, thickness=None, with_center=False)
        # Test the other parameters
        utils_object_detectors.draw_rectangle_from_bbox(input_img, bbox_without_class, color=(0, 255, 0, 255), thickness=5, with_center=True)

        # Check errors
        with self.assertRaises(ValueError):
            utils_object_detectors.draw_rectangle_from_bbox(input_img, {'y1': 2, 'x2': 4, 'y2': 4}, color=None, thickness=None, with_center=False)
        with self.assertRaises(ValueError):
            utils_object_detectors.draw_rectangle_from_bbox(input_img, {'x1': 2, 'x2': 4, 'y2': 4}, color=None, thickness=None, with_center=False)
        with self.assertRaises(ValueError):
            utils_object_detectors.draw_rectangle_from_bbox(input_img, {'x1': 2, 'y1': 2, 'y2': 4}, color=None, thickness=None, with_center=False)
        with self.assertRaises(ValueError):
            utils_object_detectors.draw_rectangle_from_bbox(input_img, {'x1': 2, 'y1': 2, 'x2': 4}, color=None, thickness=None, with_center=False)

    def test04_check_coordinates_validity(self):
        '''Test of the function utils_object_detectors.check_coordinates_validity'''

        # Set vars
        def fonction_support_args(x1, x2, y1, y2, x, y, w, h):
            pass
        def fonction_support_kwargs(x1=None, x2=None, y1=None, y2=None, x=None, y=None, w=None, h=None):
            pass
        def fonction_support_mix(x1, y1, x, w, x2=None, y2=None, y=None, h=None):
            pass

        # Nominal case
        utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 0, 1, 0, 0, 1, 1)
        utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=0, y=0, x2=1, w=1, h=1, y2=1)
        utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, 0, 1, x2=1, y2=1, y=0, h=1)

        # Check errors - args
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(-1, 1, 0, 1, 0, 0, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, -1, 0, 1, 0, 0, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, -1, 1, 0, 0, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 0, -1, 0, 0, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 0, 1, -1, 0, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 0, 1, 0, -1, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 0, 1, 0, 1, -1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 0, 1, 0, 1, 1, -1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(2, 1, 0, 1, 0, 0, 1, 1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_args)(0, 1, 2, 1, 0, 0, 1, 1)

        # Check errors - kwargs
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=-1, y1=0, x=0, y=0, x2=1, w=1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=-1, x=0, y=0, x2=1, w=1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=-1, y=0, x2=1, w=1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=0, y=-1, x2=1, w=1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=0, y=0, x2=-1, w=1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=0, y=0, x2=1, w=-1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=0, y=0, x2=1, w=1, h=-1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=0, x=0, y=0, x2=1, w=1, h=1, y2=-1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=2, y1=0, x=0, y=0, x2=1, w=1, h=1, y2=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_kwargs)(x1=0, y1=2, x=0, y=0, x2=1, w=1, h=1, y2=1)

        # Check errors - mix
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(-1, 0, 0, 1, x2=1, y2=1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, -1, 0, 1, x2=1, y2=1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, -1, 1, x2=1, y2=1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, 0, -1, x2=1, y2=1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, 0, 1, x2=-1, y2=1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, 0, 1, x2=1, y2=-1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, 0, 1, x2=1, y2=1, y=-1, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 0, 0, 1, x2=1, y2=1, y=0, h=-1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(2, 0, 0, 1, x2=1, y2=1, y=0, h=1)
        with self.assertRaises(AssertionError):
            utils_object_detectors.check_coordinates_validity(fonction_support_mix)(0, 2, 0, 1, x2=1, y2=1, y=0, h=1)

    def test05_xyxy_to_xyhw(self):
        '''Test of the function utils_object_detectors.xyxy_to_xyhw'''

        # Set vars
        x1, y1, x2, y2 = (2, 4, 10, 15)
        x, y, h, w = (2, 4, 11, 8)  # Targets

        # Nominal case
        self.assertEqual((x, y, h, w), utils_object_detectors.xyxy_to_xyhw(x1, y1, x2, y2))

    def test06_xyhw_to_xyxy(self):
        '''Test of the function utils_object_detectors.xyhw_to_xyxy'''

        # Set vars
        x, y, h, w = (2, 4, 11, 8)
        x1, y1, x2, y2 = (2, 4, 10, 15)  # Targets

        # Nominal case
        self.assertEqual((x1, y1, x2, y2), utils_object_detectors.xyhw_to_xyxy(x, y, h, w))
        self.assertEqual((x1, y1, x2, y2), utils_object_detectors.xyhw_to_xyxy(*utils_object_detectors.xyxy_to_xyhw(x1, y1, x2, y2)))  # Fonction réciproque

    def test07_xyxy_to_cxcyhw(self):
        '''Test of the function utils_object_detectors.xyxy_to_cxcyhw'''

        # Set vars
        x1, y1, x2, y2 = (2, 4, 10, 15)
        cx, cy, h, w = (6, 9.5, 11, 8)  # Targets

        # Nominal case
        self.assertEqual((cx, cy, h, w), utils_object_detectors.xyxy_to_cxcyhw(x1, y1, x2, y2))

    def test08_get_area_from_xyxy(self):
        '''Test of the function utils_object_detectors.get_area_from_xyxy'''

        # Set vars
        x1, y1, x2, y2 = (2, 4, 10, 15)
        area = 88  # Target

        # Nominal case
        self.assertEqual(area, utils_object_detectors.get_area_from_xyxy(x1, y1, x2, y2))

    def test09_get_iou(self):
        '''Test of the function utils_object_detectors.get_iou'''

        # Set vars
        coordinatesA = (2, 2, 5, 5)
        coordinatesB = (20, 20, 25, 25)
        coordinatesC = (3, 3, 7, 7)
        intersection_A_C = 2 * 2
        union_A_C = (1 * 3 + 1 * 2) + (2 * 4 + 2 * 2) + intersection_A_C  # Cf. figure below
        IOU_A_C = intersection_A_C / union_A_C

        #   2  3       5      7
        # 2  _   _   _
        #   |    _   _ | _  _
        #   |  |       |      |
        # 5 |_ | _   _ |      |
        #      |              |
        # 7    | _   _   _  _ |

        # Nominal case
        self.assertEqual(1, utils_object_detectors.get_iou(coordinatesA, coordinatesA))  # Same rectangle
        self.assertEqual(0, utils_object_detectors.get_iou(coordinatesA, coordinatesB))  # No union
        self.assertEqual(0, utils_object_detectors.get_iou(coordinatesB, coordinatesA))  # No union
        self.assertAlmostEqual(IOU_A_C, utils_object_detectors.get_iou(coordinatesA, coordinatesC))
        self.assertAlmostEqual(IOU_A_C, utils_object_detectors.get_iou(coordinatesC, coordinatesA))

    def test10_get_new_img_size_from_min_side_size(self):
        '''Test of the function utils_object_detectors.get_new_img_size_from_min_side_size'''

        # Set vars
        h1, w1, min1 = (50, 100, 150)
        h1_new, w1_new = (150, 300)
        h2, w2, min2 = (50, 100, 25)
        h2_new, w2_new = (25, 50)
        h3, w3, min3 = (100, 50, 150)
        h3_new, w3_new = (300, 150)
        h4, w4, min4 = (100, 50, 25)
        h4_new, w4_new = (50, 25)
        h5, w5, min5 = (100, 100, 200)
        h5_new, w5_new = (200, 200)

        # Nominal case
        self.assertEqual((h1_new, w1_new), utils_object_detectors.get_new_img_size_from_min_side_size(h1, w1, min1))
        self.assertEqual((h2_new, w2_new), utils_object_detectors.get_new_img_size_from_min_side_size(h2, w2, min2))
        self.assertEqual((h3_new, w3_new), utils_object_detectors.get_new_img_size_from_min_side_size(h3, w3, min3))
        self.assertEqual((h4_new, w4_new), utils_object_detectors.get_new_img_size_from_min_side_size(h4, w4, min4))
        self.assertEqual((h5_new, w5_new), utils_object_detectors.get_new_img_size_from_min_side_size(h5, w5, min5))

        # Check errors
        with self.assertRaises(ValueError):
            utils_object_detectors.get_new_img_size_from_min_side_size(0, 10, 10)
        with self.assertRaises(ValueError):
            utils_object_detectors.get_new_img_size_from_min_side_size(10, 0, 10)
        with self.assertRaises(ValueError):
            utils_object_detectors.get_new_img_size_from_min_side_size(10, 10, 0)

    def test11_get_feature_map_size(self):
        '''Test of the function utils_object_detectors.get_feature_map_size'''

        # Set vars
        h1, w1, ratio1 = (50, 100, 10)
        h1_new, w1_new = (5, 10)
        h2, w2, ratio2 = (200, 100, 25)
        h2_new, w2_new = (8, 4)
        h3, w3, ratio3 = (100, 200, 15)
        h3_new, w3_new = (6, 13)

        # Nominal case
        self.assertEqual((h1_new, w1_new), utils_object_detectors.get_feature_map_size(h1, w1, ratio1))
        self.assertEqual((h2_new, w2_new), utils_object_detectors.get_feature_map_size(h2, w2, ratio2))
        self.assertEqual((h3_new, w3_new), utils_object_detectors.get_feature_map_size(h3, w3, ratio3))

        # Check errors
        with self.assertRaises(ValueError):
            utils_object_detectors.get_feature_map_size(0, 10, 10)
        with self.assertRaises(ValueError):
            utils_object_detectors.get_feature_map_size(10, 0, 10)
        with self.assertRaises(ValueError):
            utils_object_detectors.get_feature_map_size(10, 10, 0)

    def test12_calc_regr(self):
        '''Test of the function utils_object_detectors.calc_regr'''

        # Set vars
        bbox = (5, 13, 9, 18)  # Centers 7 / 15.5, width 4, height 5
        anchor = (10, 8, 16, 20)  # Centers 13 / 14, width 6, height 12
        tx, ty = (-1., 0.125)  # -6 / 6, 1.5 / 12
        th, tw = (np.log(5/12), np.log(4/6))  # 5 / 12, 4 / 6

        # Nominal case
        self.assertEqual((tx, ty, th, tw), utils_object_detectors.calc_regr(bbox, anchor))

    def test13_apply_regression(self):
        '''Test of the function utils_object_detectors.apply_regression'''

        # Set vars
        bbox = (5, 13, 5, 4)  # x, y, h, w
        x_anc, y_anc, h_anc, w_anc = (10, 8, 12, 6)
        tx, ty = (-1., 0.125)
        th, tw = (np.log(5/12), np.log(4/6))
        coordinates_and_regression = (x_anc, y_anc, h_anc, w_anc, tx, ty, th, tw)

        # Nominal case
        self.assertEqual(bbox, utils_object_detectors.apply_regression(coordinates_and_regression))

        # Check reciprocal functions
        bbox = (5, 13, 9, 18)
        anchor = (10, 8, 16, 20)
        bbox_target = utils_object_detectors.xyxy_to_xyhw(*bbox)
        tx, ty, th, tw = utils_object_detectors.calc_regr(bbox, anchor)
        x_anc, y_anc, h_anc, w_anc = utils_object_detectors.xyxy_to_xyhw(*anchor)
        coordinates_and_regression = (x_anc, y_anc, h_anc, w_anc, tx, ty, th, tw)
        self.assertEqual(bbox_target, utils_object_detectors.apply_regression(coordinates_and_regression))

    def test14_non_max_suppression_fast(self):
        '''Test of the function utils_object_detectors.non_max_suppression_fast'''

        # Set vars
        img_boxes_coordinates = np.array([
            [2, 2, 4, 4],
            [3, 2, 4, 4],  # IOU 1/2 with first,
            [3, 2, 5, 4],  # IOU 1/3 with first, 1/2 with second
        ])

        ### Nominal case
        # Test 1
        img_boxes_probas_1 = np.array([0.9, 0.5, 0.7])
        nms_overlap_threshold_1 = 0.4
        nms_max_boxes_1 = 10
        img_boxes_classes_1 = None
        img_boxes_coordinates_res_1 = np.array([
            [2, 2, 4, 4],
            [3, 2, 5, 4],
        ])
        img_boxes_probas_res_1 = np.array([0.9, 0.7])
        img_boxes_classes_res_1 = None
        res_11, res_12, res_13 = utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_1,
                                                                                 nms_overlap_threshold_1, nms_max_boxes_1,
                                                                                 img_boxes_classes_1)
        np.testing.assert_array_equal(res_11, img_boxes_coordinates_res_1)
        np.testing.assert_array_equal(res_12, img_boxes_probas_res_1)
        np.testing.assert_array_equal(res_13, img_boxes_classes_res_1)

        # Test 2 - higher threshold  & check proba order
        img_boxes_probas_2 = np.array([0.9, 0.5, 0.7])
        nms_overlap_threshold_2 = 0.9
        nms_max_boxes_2 = 10
        img_boxes_classes_2 = None
        img_boxes_coordinates_res_2 = np.array([
            [2, 2, 4, 4],
            [3, 2, 5, 4],
            [3, 2, 4, 4],
        ])
        img_boxes_probas_res_2 = np.array([0.9, 0.7, 0.5])
        img_boxes_classes_res_2 = None
        res_21, res_22, res_23 = utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_2,
                                                                                 nms_overlap_threshold_2, nms_max_boxes_2,
                                                                                 img_boxes_classes_2)
        np.testing.assert_array_equal(res_21, img_boxes_coordinates_res_2)
        np.testing.assert_array_equal(res_22, img_boxes_probas_res_2)
        np.testing.assert_array_equal(res_23, img_boxes_classes_res_2)

        # Test 3 - lower threshold
        img_boxes_probas_3 = np.array([0.9, 0.5, 0.7])
        nms_overlap_threshold_3 = 0.1
        nms_max_boxes_3 = 10
        img_boxes_classes_3 = None
        img_boxes_coordinates_res_3 = np.array([
            [2, 2, 4, 4],
        ])
        img_boxes_probas_res_3 = np.array([0.9])
        img_boxes_classes_res_3 = None
        res_31, res_32, res_33 = utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_3,
                                                                                 nms_overlap_threshold_3, nms_max_boxes_3,
                                                                                 img_boxes_classes_3)
        np.testing.assert_array_equal(res_31, img_boxes_coordinates_res_3)
        np.testing.assert_array_equal(res_32, img_boxes_probas_res_3)
        np.testing.assert_array_equal(res_33, img_boxes_classes_res_3)

        # Test 4 - we diminish the maximal number of boxes
        img_boxes_probas_4 = np.array([0.9, 0.5, 0.7])
        nms_overlap_threshold_4 = 0.4
        nms_max_boxes_4 = 1
        img_boxes_classes_4 = None
        img_boxes_coordinates_res_4 = np.array([
            [2, 2, 4, 4],
        ])
        img_boxes_probas_res_4 = np.array([0.9])
        img_boxes_classes_res_4 = None
        res_41, res_42, res_43 = utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_4,
                                                                                 nms_overlap_threshold_4, nms_max_boxes_4,
                                                                                 img_boxes_classes_4)
        np.testing.assert_array_equal(res_41, img_boxes_coordinates_res_4)
        np.testing.assert_array_equal(res_42, img_boxes_probas_res_4)
        np.testing.assert_array_equal(res_43, img_boxes_classes_res_4)

        # Test 5 - with img_boxes_classes
        img_boxes_probas_5 = np.array([0.9, 0.5, 0.7])
        nms_overlap_threshold_5 = 0.4
        nms_max_boxes_5 = 10
        img_boxes_classes_5 = np.array(['titi', 'tata', 'toto'])
        img_boxes_coordinates_res_5 = np.array([
            [2, 2, 4, 4],
            [3, 2, 5, 4],
        ])
        img_boxes_probas_res_5 = np.array([0.9, 0.7])
        img_boxes_classes_res_5 = np.array(['titi', 'toto'])
        res_51, res_52, res_53 = utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_5,
                                                                                 nms_overlap_threshold_5, nms_max_boxes_5,
                                                                                 img_boxes_classes_5)
        np.testing.assert_array_equal(res_51, img_boxes_coordinates_res_5)
        np.testing.assert_array_equal(res_52, img_boxes_probas_res_5)
        np.testing.assert_array_equal(res_53, img_boxes_classes_res_5)

        # Test 6 - with a different proba order
        img_boxes_probas_6 = np.array([0.5, 0.9, 0.7])
        nms_overlap_threshold_6 = 0.4
        nms_max_boxes_6 = 10
        img_boxes_classes_6 = np.array(['titi', 'tata', 'toto'])
        img_boxes_coordinates_res_6 = np.array([
            [3, 2, 4, 4]
        ])
        img_boxes_probas_res_6 = np.array([0.9])
        img_boxes_classes_res_6 = np.array(['tata'])
        res_61, res_62, res_63 = utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_6,
                                                                                 nms_overlap_threshold_6, nms_max_boxes_6,
                                                                                 img_boxes_classes_6)
        np.testing.assert_array_equal(res_61, img_boxes_coordinates_res_6)
        np.testing.assert_array_equal(res_62, img_boxes_probas_res_6)
        np.testing.assert_array_equal(res_63, img_boxes_classes_res_6)

        # Check errors
        with self.assertRaises(ValueError):
            utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, np.array([0.5, 0.9]),
                                                            nms_overlap_threshold_6, nms_max_boxes_6,
                                                            img_boxes_classes_6)
        with self.assertRaises(ValueError):
            utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_6,
                                                            0, nms_max_boxes_6,
                                                            img_boxes_classes_6)
        with self.assertRaises(ValueError):
            utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_6,
                                                            1.01, nms_max_boxes_6,
                                                            img_boxes_classes_6)
        with self.assertRaises(ValueError):
            utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_6,
                                                            nms_overlap_threshold_6, 0,
                                                            img_boxes_classes_6)
        with self.assertRaises(ValueError):
            utils_object_detectors.non_max_suppression_fast(img_boxes_coordinates, img_boxes_probas_6,
                                                            nms_overlap_threshold_6, nms_max_boxes_6,
                                                            np.array(['titi', 'tata']))


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import keras
import shutil
import tensorflow
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.models_training import utils_deep_keras
from {{package_name}}.models_training.classifiers.model_cnn_classifier import ModelCnnClassifier

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)
    tensorflow.keras.backend.clear_session()

class ModelCnnClassifierTests(unittest.TestCase):
    '''Main class to test model_cnn_classifier'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_model_cnn_classifier_init(self):
        '''Test of {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier.__init__'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Init., test all parameters
        model = ModelCnnClassifier(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertEqual(model.model_type, 'classifier')
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.model, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=8)
        self.assertEqual(model.batch_size, 8)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, epochs=42)
        self.assertEqual(model.epochs, 42)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, validation_split=0.3)
        self.assertEqual(model.validation_split, 0.3)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, patience=65)
        self.assertEqual(model.patience, 65)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, width=22)
        self.assertEqual(model.width, 22)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, height=56)
        self.assertEqual(model.height, 56)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, depth=4)
        self.assertEqual(model.depth, 4)
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, color_mode='rgba')
        self.assertEqual(model.color_mode, 'rgba')
        remove_dir(model_dir)

        #
        model = ModelCnnClassifier(model_dir=model_dir, in_memory=True)
        self.assertEqual(model.in_memory, True)
        remove_dir(model_dir)

        # keras_params must accept anything !
        model = ModelCnnClassifier(model_dir=model_dir, keras_params={'toto': 5})
        self.assertEqual(model.keras_params, {'toto': 5})
        remove_dir(model_dir)

    # We do not test the fit method. It is already done in test_model_keras.py

    def test02_model_cnn_classifier_predict(self):
        '''Test of the method predict of {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'test_data_1')
        filenames = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                     'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                     'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        df_train_mono = pd.DataFrame({
            'file_class': ['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'dog', 'dog', 'dog', 'dog', 'cat', 'cat'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        df_train_multi = pd.DataFrame({
            'file_class': ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })

        # Classification - Mono-label - Mono-Class
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_mono, df_valid=df_train_mono)
        preds = model.predict(df_train_mono, return_proba=False)
        self.assertEqual(preds.shape, (df_train_mono.shape[0],))
        proba = model.predict(df_train_mono, return_proba=True)
        self.assertEqual(proba.shape, (df_train_mono.shape[0], 2)) # 2 classes
        remove_dir(model_dir)
        # in_memory
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2, in_memory=True)
        model.fit(df_train_mono, df_valid=df_train_mono)
        preds = model.predict(df_train_mono, return_proba=False)
        self.assertEqual(preds.shape, (df_train_mono.shape[0],))
        proba = model.predict(df_train_mono, return_proba=True)
        self.assertEqual(proba.shape, (df_train_mono.shape[0], 2)) # 2 classes
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_multi, df_valid=df_train_multi)
        preds = model.predict(df_train_multi, return_proba=False)
        self.assertEqual(preds.shape, (df_train_multi.shape[0],))
        proba = model.predict(df_train_multi, return_proba=True)
        self.assertEqual(proba.shape, (df_train_multi.shape[0], 3)) # 3 classes
        remove_dir(model_dir)
        # in_memory
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2, in_memory=True)
        model.fit(df_train_multi, df_valid=df_train_multi)
        preds = model.predict(df_train_multi, return_proba=False)
        self.assertEqual(preds.shape, (df_train_multi.shape[0],))
        proba = model.predict(df_train_multi, return_proba=True)
        self.assertEqual(proba.shape, (df_train_multi.shape[0], 3)) # 3 classes
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
            model.predict(df_train_multi)
        remove_dir(model_dir)

    def test03_model_cnn_classifier_predict_proba(self):
        '''Test of the method predict_proba of {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'test_data_1')
        filenames = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                     'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                     'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        df_train_mono = pd.DataFrame({
            'file_class': ['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'dog', 'dog', 'dog', 'dog', 'cat', 'cat'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        df_train_multi = pd.DataFrame({
            'file_class': ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })

        # Classification - Mono-label - Mono-Class
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_mono, df_valid=df_train_mono)
        preds = model.predict_proba(df_train_mono)
        self.assertEqual(preds.shape, (df_train_mono.shape[0], 2)) # 2 classes
        self.assertTrue(isinstance(preds[0][0], (np.floating, float)))
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_multi, df_valid=df_train_multi)
        preds = model.predict_proba(df_train_multi)
        self.assertEqual(preds.shape, (df_train_multi.shape[0], 3)) # 3 classes
        self.assertTrue(isinstance(preds[0][0], (np.floating, float)))
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
            model.predict_proba(df_train_multi)
        remove_dir(model_dir)

    def test04_model_cnn_classifier_get_predict_position(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier.get_predict_position'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'test_data_1')
        filenames = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                     'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                     'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        df_train_mono = pd.DataFrame({
            'file_class': ['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'dog', 'dog', 'dog', 'dog', 'cat', 'cat'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        y_true_mono = list(df_train_mono['file_class'].values)
        df_train_multi = pd.DataFrame({
            'file_class': ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        y_true_multi = list(df_train_multi['file_class'].values)

        # Classification - Mono-label - Mono-Class
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_mono)
        predict_positions = model.get_predict_position(df_train_mono, y_true_mono)
        self.assertEqual(predict_positions.shape, (df_train_mono.shape[0],))
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_multi)
        predict_positions = model.get_predict_position(df_train_multi, y_true_multi)
        self.assertEqual(predict_positions.shape, (df_train_multi.shape[0],))
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
            model.get_predict_position(df_train_mono, y_true_mono)
        remove_dir(model_dir)

    def test05_model_cnn_classifier_get_model(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier._get_model'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelCnnClassifier(model_dir=model_dir, epochs=2, batch_size=2)

        # Nominal case
        model.list_classes = ['a', 'b']  # We force the creation of a list of classes
        model_res = model._get_model()
        self.assertTrue(isinstance(model_res, keras.Model))

        # Clean
        remove_dir(model_dir)

    def test06_model_cnn_classifier_save(self):
        '''Test of the method save of {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelCnnClassifier(model_dir=model_dir)
        model.model_type = 'classifier' # We test classifier
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5'))) -> no model trained
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"preprocess_input.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'keras')
        self.assertTrue('batch_size' in configs.keys())
        self.assertTrue('epochs' in configs.keys())
        self.assertTrue('validation_split' in configs.keys())
        self.assertTrue('patience' in configs.keys())
        self.assertTrue('width' in configs.keys())
        self.assertTrue('height' in configs.keys())
        self.assertTrue('depth' in configs.keys())
        self.assertTrue('color_mode' in configs.keys())
        self.assertTrue('in_memory' in configs.keys())
        self.assertTrue('data_augmentation_params' in configs.keys())
        self.assertTrue('nb_train_generator_images_to_save' in configs.keys())
        self.assertTrue('keras_params' in configs.keys())
        self.assertTrue('_get_model' in configs.keys())
        self.assertTrue('_get_learning_rate_scheduler' in configs.keys())
        self.assertTrue('custom_objects' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        remove_dir(model_dir)

        # Use custom_objects containing a "partial" function
        model = ModelCnnClassifier(model_dir=model_dir)
        model.model_type = 'classifier' # We do not test 'regressor', it is the same thing
        custom_objects = utils_deep_keras.custom_objects
        custom_objects['fb_loss'] = utils_deep_keras.get_fb_loss(0.5)
        model.custom_objects = custom_objects
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5'))) -> no model trained
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"preprocess_input.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'keras')
        self.assertTrue('batch_size' in configs.keys())
        self.assertTrue('epochs' in configs.keys())
        self.assertTrue('validation_split' in configs.keys())
        self.assertTrue('patience' in configs.keys())
        self.assertTrue('width' in configs.keys())
        self.assertTrue('height' in configs.keys())
        self.assertTrue('depth' in configs.keys())
        self.assertTrue('color_mode' in configs.keys())
        self.assertTrue('in_memory' in configs.keys())
        self.assertTrue('data_augmentation_params' in configs.keys())
        self.assertTrue('nb_train_generator_images_to_save' in configs.keys())
        self.assertTrue('keras_params' in configs.keys())
        self.assertTrue('_get_model' in configs.keys())
        self.assertTrue('_get_learning_rate_scheduler' in configs.keys())
        self.assertTrue('custom_objects' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        remove_dir(model_dir)

    def test07_model_cnn_classifier_reload_model(self):
        '''Test of the method reload_model of {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'test_data_1')
        filenames = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                     'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                     'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        df_train_mono = pd.DataFrame({
            'file_class': ['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'dog', 'dog', 'dog', 'dog', 'cat', 'cat'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        df_train_multi = pd.DataFrame({
            'file_class': ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })

        # Classification - Mono-label - Mono-Class
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        fit_params = model.fit(df_train_mono, df_valid=None)
        x_batch = fit_params['x'].next()
        model.save()
        # Reload keras
        hdf5_path = os.path.join(model.model_dir, 'best.hdf5')
        reloaded_model = model.reload_model(hdf5_path)
        np.testing.assert_almost_equal([list(_) for _ in reloaded_model.predict(x_batch[0])], [list(_) for _ in model.model.predict(x_batch[0])], 3)
        # Test without custom_objects
        model.custom_objects = None
        reloaded_model = model.reload_model(hdf5_path)
        np.testing.assert_almost_equal([list(_) for _ in reloaded_model.predict(x_batch[0])], [list(_) for _ in model.model.predict(x_batch[0])], 3)
        # Clean
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        fit_params = model.fit(df_train_multi, df_valid=None)
        x_batch = fit_params['x'].next()
        model.save()
        # Reload keras
        hdf5_path = os.path.join(model.model_dir, 'best.hdf5')
        reloaded_model = model.reload_model(hdf5_path)
        np.testing.assert_almost_equal([list(_) for _ in reloaded_model.predict(x_batch[0])], [list(_) for _ in model.model.predict(x_batch[0])], 3)
        # Test without custom_objects
        model.custom_objects = None
        reloaded_model = model.reload_model(hdf5_path)
        np.testing.assert_almost_equal([list(_) for _ in reloaded_model.predict(x_batch[0])], [list(_) for _ in model.model.predict(x_batch[0])], 3)
        # Clean
        remove_dir(model_dir)

    def test08_model_cnn_classifier_reload_from_standalone(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_cnn_classifier.ModelCnnClassifier.reload_from_standalone'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'test_data_1')
        filenames = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                     'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                     'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        df_train_mono = pd.DataFrame({
            'file_class': ['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'dog', 'dog', 'dog', 'dog', 'cat', 'cat'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        df_train_multi = pd.DataFrame({
            'file_class': ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })

        ############################################
        # Classification - Mono class
        ############################################

        # Create model
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        fit_params = model.fit(df_train_mono)
        model.save()
        # Reload
        conf_path = os.path.join(model.model_dir, "configurations.json")
        hdf5_path = os.path.join(model.model_dir, "best.hdf5")
        preprocess_input_path = os.path.join(model.model_dir, "preprocess_input.pkl")
        new_model = ModelCnnClassifier()
        new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
                                         preprocess_input_path=preprocess_input_path)
        # Tests
        self.assertEqual(model.model_name, new_model.model_name)
        self.assertEqual(model.model_type, new_model.model_type)
        self.assertEqual(model.trained, new_model.trained)
        self.assertEqual(model.nb_fit, new_model.nb_fit)
        self.assertEqual(model.level_save, new_model.level_save)
        self.assertEqual(model.list_classes, new_model.list_classes)
        self.assertEqual(model.dict_classes, new_model.dict_classes)
        self.assertEqual(model.batch_size, new_model.batch_size)
        self.assertEqual(model.epochs, new_model.epochs)
        self.assertEqual(model.validation_split, new_model.validation_split)
        self.assertEqual(model.patience, new_model.patience)
        self.assertEqual(model.width, new_model.width)
        self.assertEqual(model.height, new_model.height)
        self.assertEqual(model.depth, new_model.depth)
        self.assertEqual(model.color_mode, new_model.color_mode)
        self.assertEqual(model.in_memory, new_model.in_memory)
        self.assertEqual(model.data_augmentation_params, new_model.data_augmentation_params)
        self.assertEqual(model.nb_train_generator_images_to_save, new_model.nb_train_generator_images_to_save)
        self.assertEqual(model.keras_params, new_model.keras_params)
        self.assertEqual(model.custom_objects, new_model.custom_objects)
        # We can't really test the pipeline so we test predictions
        self.assertEqual([list(_) for _ in model.predict_proba(df_train_mono)], [list(_) for _ in new_model.predict_proba(df_train_mono)])
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        ############################################
        # Classification - Multi class
        ############################################

        # Create model
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        fit_params = model.fit(df_train_multi)
        model.save()
        # Reload
        conf_path = os.path.join(model.model_dir, "configurations.json")
        hdf5_path = os.path.join(model.model_dir, "best.hdf5")
        preprocess_input_path = os.path.join(model.model_dir, "preprocess_input.pkl")
        new_model = ModelCnnClassifier()
        new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
                                         preprocess_input_path=preprocess_input_path)
        # Tests
        self.assertEqual(model.model_name, new_model.model_name)
        self.assertEqual(model.model_type, new_model.model_type)
        self.assertEqual(model.trained, new_model.trained)
        self.assertEqual(model.nb_fit, new_model.nb_fit)
        self.assertEqual(model.level_save, new_model.level_save)
        self.assertEqual(model.list_classes, new_model.list_classes)
        self.assertEqual(model.dict_classes, new_model.dict_classes)
        self.assertEqual(model.batch_size, new_model.batch_size)
        self.assertEqual(model.epochs, new_model.epochs)
        self.assertEqual(model.validation_split, new_model.validation_split)
        self.assertEqual(model.patience, new_model.patience)
        self.assertEqual(model.width, new_model.width)
        self.assertEqual(model.height, new_model.height)
        self.assertEqual(model.depth, new_model.depth)
        self.assertEqual(model.color_mode, new_model.color_mode)
        self.assertEqual(model.in_memory, new_model.in_memory)
        self.assertEqual(model.data_augmentation_params, new_model.data_augmentation_params)
        self.assertEqual(model.nb_train_generator_images_to_save, new_model.nb_train_generator_images_to_save)
        self.assertEqual(model.keras_params, new_model.keras_params)
        self.assertEqual(model.custom_objects, new_model.custom_objects)
        # We can't really test the pipeline so we test predictions
        self.assertEqual([list(_) for _ in model.predict_proba(df_train_multi)], [list(_) for _ in new_model.predict_proba(df_train_multi)])
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)


        ############################################
        # Errors
        ############################################

        with self.assertRaises(FileNotFoundError):
            new_model = ModelCnnClassifier()
            new_model.reload_from_standalone(configuration_path='toto.json', hdf5_path=hdf5_path,
                                             preprocess_input_path=preprocess_input_path)
        with self.assertRaises(FileNotFoundError):
            new_model = ModelCnnClassifier()
            new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path='toto.pkl',
                                             preprocess_input_path=preprocess_input_path)
        with self.assertRaises(FileNotFoundError):
            new_model = ModelCnnClassifier()
            new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
                                             preprocess_input_path='toto.pkl')

        # Clean
        remove_dir(model_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import shutil
import mlflow
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.monitoring.model_logger import ModelLogger, is_running, is_local, is_mlflow_up

# Disable logging
import logging
logging.disable(logging.CRITICAL)


class ModelLoggerTests(unittest.TestCase):
    '''Main class to test model_logger'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_is_running(self):
        '''Test of the method {{package_name}}.monitoring.model_logger.is_running'''
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir)) # We just want the logger
        bad_host = 'http://toto.titi.tata.test'
        bad_port = 80
        self.assertFalse(is_running(bad_host, bad_port, model.logger))
        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test02_is_local(self):
        '''Test of the method {{package_name}}.monitoring.model_logger.is_local'''
        local = 'ceci/est/un/test'
        distant = 'http://ceci.est.un.faux.site.com'
        self.assertTrue(is_local(local))
        self.assertFalse(is_local(distant))

    def test03_model_logger_init(self):
        '''Test of the initialization of {{package_name}}.monitoring.model_logger.ModelLogger'''
        # We test with a fake host
        host = 'http://toto.titi.tata.test'
        name = 'test'
        model = ModelLogger(tracking_uri=host, experiment_name=name)
        self.assertEqual(model.tracking_uri, host)
        self.assertEqual(model.experiment_name, name)
        self.assertFalse(model.running)
        # Clear
        model.stop_run()

        # We test with nothing (by default, mlflow saves in a folder 'ml_runs')
        save_dir = os.path.join(os.getcwd(), 'mlruns')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        self.assertFalse(os.path.exists(save_dir))
        model = ModelLogger()
        self.assertEqual(model.tracking_uri, '')
        self.assertTrue(model.running)
        self.assertTrue(os.path.exists(save_dir))
        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test04_model_logger_stop_run(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.stop_run'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))
        # We activate a run via a log
        model.log_param('test', 'toto')
        # Use of stop_run
        model.stop_run()
        # Check
        self.assertEqual(mlflow.active_run(), None)
        # Clear
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test05_model_logger_log_metric(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.log_metric'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))

        # Nominal case
        model.log_metric('test', 5)
        model.log_metric('test', 5, step=2)

        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test06_model_logger_log_metrics(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.log_metrics'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))

        # Nominal case
        model.log_metrics({'test': 5})
        model.log_metrics({'test': 5}, step=2)

        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test07_model_logger_log_param(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.log_param'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))

        # Nominal case
        model.log_param('test', 5)

        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test08_model_logger_log_params(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.log_params'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))

        # Nominal case
        model.log_params({'test': 5})

        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test09_model_logger_set_tag(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.set_tag'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))

        # Nominal case
        model.set_tag('test', 5)

        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)

    def test10_model_logger_set_tags(self):
        '''Test of {{package_name}}.monitoring.model_logger.ModelLogger.set_tags'''
        # Init. logger
        save_dir = os.path.join(os.getcwd(), 'ml_flow_test')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        model = ModelLogger(os.path.relpath(save_dir))

        # Nominal case
        model.set_tags({'test': 5})

        # Clear
        model.stop_run()
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Definition of the parent class for the models
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelClass -> Parent class for the models


import os
import json
import logging
import numpy as np
import pandas as pd
import dill as pickle
from datetime import datetime
from typing import List, Union, Tuple

from sklearn.compose import ColumnTransformer
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.models_training import utils_models
from {{package_name}}.monitoring.model_logger import ModelLogger


class ModelClass:
    '''Parent class for the models'''

    _default_name = 'none'
    # Variable annotation : https://www.python.org/dev/peps/pep-0526/
    # Solves lots of typing errors, cf mypy
    multi_label: Union[bool, None]
    list_classes: Union[list, None]
    dict_classes: Union[dict, None]

    # Not implemented :
    # -> fit
    # -> predict
    # -> predict_proba
    # -> inverse_transform
    # -> get_and_save_metrics

    def __init__(self, model_dir: Union[str, None] = None, model_name: Union[str, None] = None,
                 x_col: Union[list, None] = None, y_col: Union[str, int, list, None] = None,
                 preprocess_pipeline: Union[ColumnTransformer, None] = None, level_save: str = 'HIGH') -> None:
        '''Initialization of the parent class.

        Kwargs:
            model_dir (str): Folder where to save the model
                If None, creates a directory based on the model's name and the date (most common usage)
            model_name (str): The name of the model
            x_col (list): Names of the columns used for the training - x
            y_col (str or int or list if multi-labels): Name of the model's target column(s) - y
            preprocess_pipeline (ColumnTransformer): The pipeline used for preprocessing. If None -> no preprocessing !
            level_save (str): Level of saving
                LOW: stats + configurations + logger keras - /!\\ The model can't be reused /!\\ -
                MEDIUM: LOW + hdf5 + pkl + plots
                HIGH: MEDIUM + predictions
        Raises:
            ValueError: If the object level_save is not a valid option (['LOW', 'MEDIUM', 'HIGH'])
            NotADirectoryError: If a provided model directory is not a directory (i.e. it's a file)
        '''
        if level_save not in ['LOW', 'MEDIUM', 'HIGH']:
            raise ValueError(f"The object level_save ({level_save}) is not a valid option (['LOW', 'MEDIUM', 'HIGH'])")

        # Get logger
        self.logger = logging.getLogger(__name__)

        # Model type -> 'classifier' or 'regressor' depending on the model
        self.model_type = None

        # Model name
        if model_name is None:
            self.model_name = self._default_name
        else:
            self.model_name = model_name

        # Names of the columns used
        self.x_col = x_col
        self.y_col = y_col
        # Can be None if reloading a model
        if x_col is None:
            self.logger.warning("Warning, the attribute x_col is not given! The model might not work as intended.")
        if y_col is None:
            self.logger.warning("Warning, the attribute y_col is not given! The model might not work as intended.")

        # Model folder
        if model_dir is None:
            self.model_dir = self._get_model_dir()
        else:
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
            if not os.path.isdir(model_dir):
                raise NotADirectoryError(f"{model_dir} is not a valid directory")
            self.model_dir = os.path.abspath(model_dir)

        # Preprocessing pipeline
        self.preprocess_pipeline = preprocess_pipeline
        if self.preprocess_pipeline is not None:
            try:
                check_is_fitted(self.preprocess_pipeline)
            except NotFittedError as e:
                self.logger.error("The preprocessing pipeline hasn't been fitted !")
                self.logger.error(repr(e))
                raise NotFittedError()
            # We get the associated columns (and a check if there has been a fit is done)
            self.columns_in, self.mandatory_columns = utils_models.get_columns_pipeline(self.preprocess_pipeline)
        else:
            # We can't define a "no_preprocess" pipeline since we should fit it
            # So we take care of that at the first fit
            self.logger.warning("Warning, no preprocessing pipeline given !")
            self.columns_in, self.mandatory_columns = None, None

        # Other options
        self.level_save = level_save

        # is trained ?
        self.trained = False
        self.nb_fit = 0

    def fit(self, x_train, y_train, **kwargs) -> None:
        '''Trains the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train (?): Array-like, shape = [n_samples, n_targets]
        '''
        raise NotImplementedError("'fit' needs to be overridden")

    def predict(self, x_test: pd.DataFrame, **kwargs) -> np.ndarray:
        '''Predictions on the test set

        Args:
            x_test (pd.DataFrame): DataFrame with the test data to be predicted
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        raise NotImplementedError("'predict' needs to be overridden")

    def predict_proba(self, x_test: pd.DataFrame, **kwargs) -> np.ndarray:
        '''Probabilities predicted on the test set

        Args:
            x_test (pd.DataFrame): DataFrame with the test data to be predicted
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        raise NotImplementedError("'predict_proba' needs to be overridden")

    def inverse_transform(self, y: Union[list, np.ndarray]) -> Union[list, tuple]:
        '''Gets the final format of prediction
            - Classification : classes from predictions
            - Regression : values (identity function)

        Args:
            y (list | np.ndarray): Array-like, shape = [n_samples,]
                   OR 1D array shape = [n_classes] (only one prediction)
        Returns:
            (?): Array, shape = [n_samples, ?]
        '''
        raise NotImplementedError("'inverse_transform' needs to be overridden")

    def get_and_save_metrics(self, y_true, y_pred, df_x: Union[pd.DataFrame, None] = None,
                             series_to_add: Union[List[pd.Series], None] = None, type_data: str = '',
                             model_logger: Union[ModelLogger, None] = None) -> pd.DataFrame:
        '''Gets and saves the metrics of a model

        Args:
            y_true (?): Array-like, shape = [n_samples, n_targets]
            y_pred (?): Array-like, shape = [n_samples, n_targets]
        Kwargs:
            df_x (pd.DataFrame or None): Input dataFrame used for the prediction
            series_to_add (list): List of pd.Series to add to the dataframe
            type_data (str): Type of dataset (validation, test, ...)
            model_logger (ModelLogger): Custom class to log the metrics with MLflow
        Returns:
            pd.DataFrame: The dataframe containing the statistics
        '''
        raise NotImplementedError("'get_and_save_metrics' needs to be overridden")

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''

        # Manage paths
        pkl_path = os.path.join(self.model_dir, f"{self.model_name}.pkl")
        preprocess_pipeline_path = os.path.join(self.model_dir, "preprocess_pipeline.pkl")
        conf_path = os.path.join(self.model_dir, "configurations.json")

        # Save the model & preprocessing pipeline if level_save > 'LOW'
        if self.level_save in ['MEDIUM', 'HIGH']:
            with open(pkl_path, 'wb') as f:
                pickle.dump(self, f)
            # Useful for reload_from_standalone, otherwise, saved as a class attribute
            with open(preprocess_pipeline_path, 'wb') as f:
                pickle.dump(self.preprocess_pipeline, f)

        # Saving JSON configuration
        json_dict = {
            'mainteners': 'Agence DataServices',
            'date': datetime.now().strftime("%d/%m/%Y - %H:%M:%S"),  # Not the same as the folder's name
            'package_version': utils.get_package_version(),
            'model_name': self.model_name,
            'model_dir': self.model_dir,
            'model_type': self.model_type,
            'trained': self.trained,
            'nb_fit': self.nb_fit,
            'x_col': self.x_col,
            'y_col': self.y_col,
            'columns_in': self.columns_in,
            'mandatory_columns': self.mandatory_columns,
            'level_save': self.level_save,
            'librairie': None,
        }
        # Merge json_data if not None
        if json_data is not None:
            # Priority given to json_data !
            json_dict = {**json_dict, **json_data}

        # Save conf
        with open(conf_path, 'w', encoding='{{default_encoding}}') as json_file:
            json.dump(json_dict, json_file, indent=4, cls=utils.NpEncoder)

        # Now, save a proprietes file for the model upload
        self._save_upload_properties(json_dict)

    def _save_upload_properties(self, json_dict: Union[dict, None] = None) -> None:
        '''Prepares a configuration file for a future export (e.g on an artifactory)

        Kwargs:
            json_dict: Configurations to save
        '''
        if json_dict is None:
            json_dict = {}

        # Manage paths
        proprietes_path = os.path.join(self.model_dir, "proprietes.json")
        vanilla_model_upload_instructions = os.path.join(utils.get_ressources_path(), 'model_upload_instructions.md')
        specific_model_upload_instructions = os.path.join(self.model_dir, "model_upload_instructions.md")

        # First, we define a list of "allowed" properties
        allowed_properties = ["mainteners", "date", "package_version", "model_name", "list_classes",
                              "librairie", "fit_time"]
        # Now we filter these properties
        final_dict = {k: v for k, v in json_dict.items() if k in allowed_properties}
        # Save
        with open(proprietes_path, 'w', encoding='{{default_encoding}}') as f:
            json.dump(final_dict, f, indent=4, cls=utils.NpEncoder)

        # Add instructions to upload a model to a storage solution (e.g. Artifactory)
        with open(vanilla_model_upload_instructions, 'r', encoding='{{default_encoding}}') as f:
            content = f.read()
        # TODO: to be improved
        new_content = content.replace('model_dir_path_identifier', os.path.abspath(self.model_dir))
        with open(specific_model_upload_instructions, 'w', encoding='{{default_encoding}}') as f:
            f.write(new_content)

    def _get_model_dir(self) -> str:
        '''Gets a folder where to save the model

        Returns:
            str: Path to the folder
        '''
        models_dir = utils.get_models_path()
        subfolder = os.path.join(models_dir, self.model_name)
        folder_name = datetime.now().strftime(f"{self.model_name}_%Y_%m_%d-%H_%M_%S")
        model_dir = os.path.join(subfolder, folder_name)
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)
        return model_dir

    def _check_input_format(self, x_input: Union[pd.DataFrame, np.ndarray], y_input: Union[pd.DataFrame, pd.Series, np.ndarray, None] = None,
                            fit_function: bool = False) -> Tuple[Union[pd.DataFrame, np.ndarray], Union[pd.DataFrame, pd.Series, np.ndarray, None] ]:
        '''Checks the inputs of a function. We check the number of columns and the ordering.
        Warnings if :
            - Not the right columns
            - Columns not in the right order
        If fit and x_col and/or y_col is not defined -> warning + use the input columns
        We also set the pipeline, columns_in and mandatory_columns if equal to None

        Args:
            x_input (pd.DataFrame, np.ndarray): Array-like, shape = [n_samples, n_features]
        Kwargs:
            y_input (pd.DataFrame, pd.Series, np.ndarray): Array-like, shape = [n_samples, n_target]
                Mandatory if fit_function
            fit_function (bool): If it is a fit function
        Raises:
            AttributeError: If fit_function == True, but y_input is None
            ValueError: If one of the inputs hasn't the right number of columns
        Returns:
            (pd.DataFrame, np.ndarray): x_input, may be reordered if needed
            (pd.DataFrame, pd.Series, np.ndarray): y_input, may be reordered if needed
        '''
        # Getting some info first
        x_input_shape = x_input.shape[-1] if len(x_input.shape) > 1 else 1
        if y_input is not None:
            y_input_shape = y_input.shape[-1] if len(y_input.shape) > 1 else 1
        else:
            y_input_shape = 0 # non utilisé

        # Manage fit_function = True
        if fit_function == True:
            if y_input is None:
                raise AttributeError("The argument y_input is mandatory if fit_function == True")
            if self.x_col is None:
                self.logger.warning("Warning, the attribute x_col was not given when creating the model")
                self.logger.warning("We set it now with the input data of the fit function")
                if hasattr(x_input, 'columns'):
                    self.x_col = list(x_input.columns)
                else:
                    self.x_col = [_ for _ in range(x_input_shape)]
            # Same thing for y_col
            if self.y_col is None:
                self.logger.warning("Warning, the attribute y_col was not given when creating the model")
                self.logger.warning("We set it now with the input data of the fit function")
                if hasattr(y_input, 'columns'):
                    self.y_col = list(y_input.columns)
                else:
                    self.y_col = [_ for _ in range(y_input_shape)]
                # If there is only one element, we get rid of the list
                if y_input_shape == 1:
                    self.y_col = self.y_col[0]
            # If pipeline, columns_in or mandatory_columns is None, sets it
            if self.preprocess_pipeline is None: # ie no pipeline given when initializing the class
                preprocess_str = "no_preprocess"
                preprocess_pipeline = preprocess.get_pipeline(preprocess_str) # Warning, the pipeline must be fitted
                preprocess_pipeline.fit(x_input) # We fit the pipeline to set the necessary columns for the pipeline
                self.preprocess_pipeline = preprocess_pipeline
                self.columns_in, self.mandatory_columns = utils_models.get_columns_pipeline(self.preprocess_pipeline)

        # Checking x_input
        if self.x_col is None:
            self.logger.warning("Can't check the input format (x) because x_col is not set...")
        else:
            # Checking x_input format
            x_col_len = len(self.x_col)
            if x_input_shape != x_col_len:
                raise ValueError(f"Input data (x) is not in the right format ({x_input_shape} != {x_col_len})")
            # We check the presence of the columns
            if hasattr(x_input, 'columns'):
                can_reorder = True
                for col in self.x_col:
                    if col not in x_input.columns:
                        can_reorder = False
                        self.logger.warning(f"The column {col} is missing from the input (x)")
                # If we can't reorder, we write a warning message, otherwise we check if it is needed
                if not can_reorder:
                    self.logger.warning("The names of the columns do not match. The process continues since there is the right number of columns")
                else:
                    if list(x_input.columns) != self.x_col:
                        self.logger.warning("The input columns (x) are not in the right order -> automatic reordering !")
                        x_input = x_input[self.x_col]
            else:
                self.logger.warning(f"The input (x) does not have the 'columns' attribute -> can't check the ordering of the columns")

        # Checking y_input
        if y_input is not None:
            if self.y_col is None:
                self.logger.warning("Can't check the input format (y) because y_col is not set...")
            else:
                # Checking y_input format
                y_col_len = len(self.y_col) if type(self.y_col) == list else 1
                if y_input_shape != y_col_len:
                    raise ValueError(f"Input data (y) is not in the right format ({y_input_shape} != {y_col_len})")
                # We check the presence of the columns
                if hasattr(y_input, 'columns'):
                    can_reorder = True
                    target_cols = self.y_col if type(self.y_col) == list else [self.y_col]
                    for col in target_cols:
                        if col not in y_input.columns:
                            can_reorder = False
                            self.logger.warning(f"The column {col} is missing from the input (y)")
                    # If can't reorder we write a warning message, otherwise we check if it is needed
                    if not can_reorder:
                        self.logger.warning("The names of the columns do not match. The process continues since there is the right number of columns")
                    else:
                        if list(y_input.columns) != target_cols:
                            self.logger.warning("The input columns (y) are not in the right order -> automatic reordering !")
                            y_input = y_input[target_cols]
                else:
                    self.logger.warning(f"The input (y) does not have the 'columns' attribute -> can't check the ordering of the columns")

        # Return
        return x_input, y_input

    def display_if_gpu_activated(self) -> None:
        '''Displays if a GPU is being used'''
        if self._is_gpu_activated():
            ascii_art = '''
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*         (=========)                                                                                                            (=========)         *
*         |=========|                                                                                                            |=========|         *
*         |====_====|                                                                                                            |====_====|         *
*         |== / \ ==|                                                                                                            |== / \ ==|         *
*         |= / _ \ =|                                                                                                            |= / _ \ =|         *
*      _  |=| ( ) |=|                                                                                                         _  |=| ( ) |=|         *
*     /=\ |=|     |=| /=\                                                                                                    /=\ |=|     |=| /=\     *
*     |=| |=| GPU |=| |=|        _____ _____  _    _            _____ _______ _______      __  _______ ______ _____          |=| |=| GPU |=| |=|     *
*     |=| |=|  _  |=| |=|       / ____|  __ \| |  | |     /\   / ____|__   __|_   _\ \    / /\|__   __|  ____|  __ \         |=| |=|  _  |=| |=|     *
*     |=| |=| | | |=| |=|      | |  __| |__) | |  | |    /  \ | |       | |    | |  \ \  / /  \  | |  | |__  | |  | |        |=| |=| | | |=| |=|     *
*     |=| |=| | | |=| |=|      | | |_ |  ___/| |  | |   / /\ \| |       | |    | |   \ \/ / /\ \ | |  |  __| | |  | |        |=| |=| | | |=| |=|     *
*     |=| |=| | | |=| |=|      | |__| | |    | |__| |  / ____ \ |____   | |   _| |_   \  / ____ \| |  | |____| |__| |        |=| |=| | | |=| |=|     *
*     |=| |/  | |  \| |=|       \_____|_|     \____/  /_/    \_\_____|  |_|  |_____|   \/_/    \_\_|  |______|_____/         |=| |/  | |  \| |=|     *
*     |=|/    | |    \|=|                                                                                                    |=|/    | |    \|=|     *
*     |=/ ADS |_| ADS \=|                                                                                                    |=/ ADS |_| ADS \=|     *
*     |(_______________)|                                                                                                    |(_______________)|     *
*     |=| |_|__|__|_| |=|                                                                                                    |=| |_|__|__|_| |=|     *
*     |=|   ( ) ( )   |=|                                                                                                    |=|   ( ) ( )   |=|     *
*    /===\           /===\                                                                                                  /===\           /===\    *
*   |||||||         |||||||                                                                                                |||||||         |||||||   *
*   -------         -------                                                                                                -------         -------   *
*    (~~~)           (~~~)                                                                                                  (~~~)           (~~~)    *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
            '''
        else:
            ascii_art = ''
        print(ascii_art)

    def _is_gpu_activated(self) -> bool:
        '''Checks if we use a GPU

        Returns:
            bool: whether GPU is available or not
        '''
        # By default, no GPU
        return False

if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import numpy as np
import pandas as pd

from {{package_name}} import utils

# Disable logging
import logging
logging.disable(logging.CRITICAL)


class UtilsTests(unittest.TestCase):
    '''Main class to test all functions in utils.py'''
    # We avoid tqdm prints
    pd.Series.progress_apply = pd.Series.apply

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_read_folder_classification(self):
        '''Test of the function utils.read_folder_classification'''
        # We set DIR_PATH equal to the test directory in order to use test data
        current_dir = os.path.abspath(os.getcwd())
        test_data_path = os.path.join(current_dir, 'test_data')
        utils.DIR_PATH = test_data_path

        #######################################
        # Format 1 : metadata file
        #######################################
        sorted_file_list = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                            'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                            'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        sorted_classes_list = ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay']

        # Nominal case
        directory_path = os.path.join(utils.get_data_path(), 'test_data_1')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_1', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'no_preprocess')

        # With different sep
        directory_path = os.path.join(utils.get_data_path(), 'test_data_1_diff_sep')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path, sep=',')
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_1_diff_sep', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'no_preprocess')

        # With preprocessing
        directory_path = os.path.join(utils.get_data_path(), 'test_data_1_with_preprocess')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_1_with_preprocess', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'test_preprocess')

        # Without classes
        directory_path = os.path.join(utils.get_data_path(), 'test_data_1_no_class')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path, accept_no_metadata=True)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_1_no_class', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual(classes_list, None)
        self.assertEqual(preprocess_str, 'no_preprocess')

        # Check errors
        directory_path = os.path.join(utils.get_data_path(), 'bad_test_data_1')
        with self.assertRaises(ValueError):
            utils.read_folder_classification(directory_path)  # No filename column
        directory_path = os.path.join(utils.get_data_path(), 'test_data_1_no_class')
        with self.assertRaises(ValueError):
            path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path, accept_no_metadata=False)

        #######################################
        # Format 2 : prefix class
        #######################################
        sorted_file_list = ['birman_Birman_1.jpg', 'birman_Birman_2.jpg', 'birman_Birman_36.jpg', 'birman_Birman_4.jpg',
                            'bombay_Bombay_1.png', 'bombay_Bombay_10.jpg', 'bombay_Bombay_19.jpg', 'bombay_Bombay_3.jpg',
                            'shiba_shiba_inu_15.jpg', 'shiba_shiba_inu_30.jpg', 'shiba_shiba_inu_31.jpg', 'shiba_shiba_inu_34.jpg']
        sorted_file_list_no_class = ['Birman1.jpg', 'Birman2.jpg', 'Birman36.jpg', 'Birman4.jpg', 'Bombay1.png', 'Bombay10.jpg',
                                     'Bombay19.jpg', 'Bombay3.jpg', 'shiba15.jpg', 'shiba30.jpg', 'shiba31.jpg', 'shiba34.jpg']
        sorted_classes_list = ['birman', 'birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba']

        # Nominal case
        directory_path = os.path.join(utils.get_data_path(), 'test_data_2')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_2', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'no_preprocess')

        # With preprocessing
        directory_path = os.path.join(utils.get_data_path(), 'test_data_2_with_preprocess')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_2_with_preprocess', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'test_preprocess')

        # Without classes
        directory_path = os.path.join(utils.get_data_path(), 'test_data_2_no_class')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_2_no_class', f) for f in sorted_file_list_no_class]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual(classes_list, None)
        self.assertEqual(preprocess_str, 'no_preprocess')

        # Check errors
        directory_path = os.path.join(utils.get_data_path(), 'bad_test_data_2')
        with self.assertRaises(RuntimeError):
            utils.read_folder_classification(directory_path)  # Not only images

        #######################################
        # Format 3 : one sub-folder per class
        #######################################
        sorted_file_list = [os.path.join('birman', 'Birman_1.jpg'), os.path.join('birman', 'Birman_2.jpg'), os.path.join('birman', 'Birman_36.jpg'),
                            os.path.join('birman', 'Birman_4.jpg'), os.path.join('bombay', 'Bombay_1.png'), os.path.join('bombay', 'Bombay_10.jpg'),
                            os.path.join('bombay', 'Bombay_19.jpg'), os.path.join('bombay', 'Bombay_3.jpg'), os.path.join('shiba', 'shiba_inu_15.jpg'),
                            os.path.join('shiba', 'shiba_inu_30.jpg'), os.path.join('shiba', 'shiba_inu_31.jpg'), os.path.join('shiba', 'shiba_inu_34.jpg')]
        sorted_classes_list = ['birman', 'birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba']

        # Nominal case
        directory_path = os.path.join(utils.get_data_path(), 'test_data_3')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_3', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'no_preprocess')

        # With preprocessing
        directory_path = os.path.join(utils.get_data_path(), 'test_data_3_with_preprocess')
        path_list, classes_list, preprocess_str = utils.read_folder_classification(directory_path)
        wanted_file_list = [os.path.join(test_data_path, '{{package_name}}-data', 'test_data_3_with_preprocess', f) for f in sorted_file_list]
        self.assertEqual(sorted(path_list), wanted_file_list)
        self.assertEqual([cl for p, cl in sorted(zip(path_list, classes_list), key=lambda pair: pair[0])], sorted_classes_list)
        self.assertEqual(preprocess_str, 'test_preprocess')

        # Check errors
        directory_path = os.path.join(utils.get_data_path(), 'bad_test_data_3')
        with self.assertRaises(RuntimeError):
            utils.read_folder_classification(directory_path)  # Not only sub-folders
        directory_path = os.path.join(utils.get_data_path(), 'bad_test_data_3_v2')
        with self.assertRaises(RuntimeError):
            utils.read_folder_classification(directory_path)  # Not only images in sub-folders

        # Clean
        utils.DIR_PATH = None

    @patch('logging.Logger._log')
    def test02_display_shape(self, PrintMockLog):
        '''Test of the function utils.display_shape'''
        # Enable the logger again
        logging.disable(logging.NOTSET)

        # Valids to test
        input_test = pd.DataFrame({'col1': ['A', 'A', 'D', np.nan, 'C', 'B'],
                                   'col2': [0, 0, 1, 2, 3, 1],
                                   'col3': [0, 0, 3, np.nan, np.nan, 3]})

        # Assert info called 1 time
        utils.display_shape(input_test)
        self.assertEqual(len(PrintMockLog.mock_calls), 1)

        # RESET DEFAULT
        logging.disable(logging.CRITICAL)

    def test03_get_chunk_limits(self):
        '''Test of the function utils.get_chunk_limits'''
        # Valids to test
        input_test_1 = pd.DataFrame()
        expected_result_1 = [(0, 0)]
        input_test_2 = pd.DataFrame([0])
        expected_result_2 = [(0, 1)]
        input_test_3 = pd.DataFrame([0] * 100000)
        expected_result_3 = [(0, 10000), (10000, 20000), (20000, 30000), (30000, 40000),
                             (40000, 50000), (50000, 60000), (60000, 70000), (70000, 80000),
                             (80000, 90000), (90000, 100000)]
        input_test_4 = pd.DataFrame([0] * 100001)
        expected_result_4 = [(0, 10000), (10000, 20000), (20000, 30000), (30000, 40000),
                             (40000, 50000), (50000, 60000), (60000, 70000), (70000, 80000),
                             (80000, 90000), (90000, 100000), (100000, 100001)]
        input_test_5 = pd.DataFrame([0] * 100)
        expected_result_5 = [(0, 10), (10, 20), (20, 30), (30, 40), (40, 50), (50, 60),
                             (60, 70), (70, 80), (80, 90), (90, 100)]

        # Nominal case
        self.assertEqual(utils.get_chunk_limits(input_test_1), expected_result_1)
        self.assertEqual(utils.get_chunk_limits(input_test_2), expected_result_2)
        self.assertEqual(utils.get_chunk_limits(input_test_3), expected_result_3)
        self.assertEqual(utils.get_chunk_limits(input_test_4), expected_result_4)
        self.assertEqual(utils.get_chunk_limits(input_test_5, chunksize=10), expected_result_5)

        # Check the input(s) type(s)
        with self.assertRaises(ValueError):
            utils.get_chunk_limits(input_test_1, chunksize=-1)

    def test04_get_configs_path(self):
        '''Test of the function utils.get_configs_path'''
        # Nominal case
        path = utils.get_configs_path()
        self.assertEqual(os.path.isdir(path), True)
        self.assertEqual(path.endswith('configs'), True)

    def test05_get_data_path(self):
        '''Test of the function utils.get_data_path'''
        # Nominal case
        path = utils.get_data_path()
        self.assertEqual(os.path.isdir(path), True)
        self.assertEqual(path.endswith('{{package_name}}-data'), True)

        # With a DIR_PATH != None
        current_dir = os.path.abspath(os.getcwd())
        utils.DIR_PATH = current_dir
        path = utils.get_data_path()
        self.assertEqual(os.path.isdir(path), True)
        self.assertEqual(path, os.path.join(current_dir, '{{package_name}}-data'))
        utils.DIR_PATH = None

    def test06_get_models_path(self):
        '''Test of the function utils.get_models_path'''
        # Nominal case
        path = utils.get_models_path()
        self.assertEqual(os.path.isdir(path), True)
        self.assertEqual(path.endswith('{{package_name}}-models'), True)

        # With a DIR_PATH != None
        current_dir = os.path.abspath(os.getcwd())
        utils.DIR_PATH = current_dir
        path = utils.get_models_path()
        self.assertEqual(os.path.isdir(path), True)
        self.assertEqual(path, os.path.join(current_dir, '{{package_name}}-models'))
        utils.DIR_PATH = None

    def test07_get_ressources_path(self):
        '''Test of the function utils.get_ressources_path'''
        # Nominal case
        path = utils.get_ressources_path()
        self.assertEqual(os.path.isdir(path), True)
        self.assertEqual(path.endswith('{{package_name}}-ressources'), True)

    def test08_get_package_version(self):
        '''Test of the function utils.get_package_version'''
        # Nominal case
        version = utils.get_package_version()
        self.assertEqual(type(version), str)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()
